@@ -44,7 +44,7 @@ names and secondary unit designators.
 Most Geo::StreetAddress::US methods return a reference to a hash containing
 address or intersection information. This
 "address specifier" hash may contain any of the following fields for a
-given address. If a given field is not present in the address, the 
+given address. If a given field is not present in the address, the
 corresponding key will be set to C<undef> in the hash.
 
 Future versions of this module may add extra fields.
@@ -67,7 +67,7 @@ Name of the street, without directional or type qualifiers.
 =head2 type
 
 Abbreviated street type, e.g. Rd, St, Ave, etc. See the USPS official
-type abbreviations at L<http://www.usps.com/ncsc/lookups/abbr_suffix.txt> 
+type abbreviations at L<http://pe.usps.com/text/pub28/pub28apc.html>
 for a list of abbreviations used.
 
 =head2 suffix
@@ -81,8 +81,7 @@ Name of the city, town, or other locale that the address is situated in.
 =head2 state
 
 The state which the address is situated in, given as its two-letter
-postal abbreviation. See L<http://www.usps.com/ncsc/lookups/abbr_state.txt>
-for a list of abbreviations used.
+postal abbreviation.  for a list of abbreviations used.
 
 =head2 zip
 
@@ -128,13 +127,13 @@ State abbreviation, as above.
 
 Five digit ZIP code, as above.
 
-=cut 
+=cut
 
 use 5.008_001;
 use strict;
 use warnings;
 
-our $VERSION = '0.99';
+our $VERSION = '1.02';
 
 =head1 GLOBAL VARIABLES
 
@@ -143,6 +142,8 @@ uses to recognize different bits of US street addresses. Although you
 will probably not need them, they are documented here for completeness's
 sake.
 
+=cut
+
 =head2 %Directional
 
 Maps directional names (north, northeast, etc.) to abbreviations (N, NE, etc.).
@@ -164,7 +165,7 @@ our %Directional = (
     northwest	=> "NW",
 );
 
-our %Direction_Code = reverse %Directional;
+our %Direction_Code; # setup in init();
 
 =head2 %Street_Type
 
@@ -539,7 +540,7 @@ our %Street_Type = (
     wy		=> "way",
 );
 
-our %_Street_Type_List = map { $_ => 1 } %Street_Type; 
+our %_Street_Type_List;	# set up in init() later;
 
 =head2 %State_Code
 
@@ -677,7 +678,7 @@ our %State_FIPS = (
     "78" => "VI",
 );
 
-our %FIPS_State = reverse %State_FIPS;
+our %FIPS_State; # setup in init() later;
 
 =head2 %Addr_Match
 
@@ -691,141 +692,9 @@ subtle ways between releases.
 
 =cut
 
-our %Addr_Match = (
-    type    => join("|", keys %_Street_Type_List),
-    number  => qr/\d+-?\d*/,
-    fraction => qr{\d+\/\d+},
-    state   => join("|",
-        # escape spaces in state names (e.g., "new york" --> "new\\ york")
-        # so they still match in the x environment below
-        map { ( quotemeta $_) } keys %State_Code, values %State_Code),
-    direct  => join("|",
-		    # map direction names to direction codes
-                    keys %Directional,
-		    # also map the dotted version of the code to the code itself
-                    map { my $c = $_;
-                          $c =~ s/(\w)/$1./g;
-                          ( quotemeta $c, $_ ) }
-                    sort { length $b <=> length $a }
-                    values %Directional),
-    dircode => join("|", keys %Direction_Code), 
-    zip	    => qr/\d{5}(?:-?\d{4})?/,  # XXX add \b?
-    corner  => qr/(?:\band\b|\bat\b|&|\@)/i,
-);
+our %Addr_Match; # setup in init()
 
-{
-    use re 'eval';
-
-    # note that expressions like [^,]+ may scan more than you expect
-    $Addr_Match{street} = qr/
-        (?:
-          # special case for addresses like 100 South Street
-          (?:($Addr_Match{direct})\W+           (?{ $_{street} = $^N })
-             ($Addr_Match{type})\b              (?{ $_{type}   = $^N }))
-          |
-          (?:($Addr_Match{direct})\W+		(?{ $_{prefix} = $^N }))?
-          (?:
-            ([^,]+)				(?{ $_{street} = $^N })
-            (?:[^\w,]+($Addr_Match{type})\b	(?{ $_{type}   = $^N }))
-            (?:[^\w,]+($Addr_Match{direct})\b	(?{ $_{suffix} = $^N }))?
-           |
-            ([^,]*\d)				(?{ $_{street} = $^N })
-            ($Addr_Match{direct})\b		(?{ $_{suffix} = $^N; $_{type}||='' })
-           |
-            ([^,]+?)				(?{ $_{street} = $^N; $_{type}||='' })
-            (?:[^\w,]+($Addr_Match{type})\b	(?{ $_{type}   = $^N }))?
-            (?:[^\w,]+($Addr_Match{direct})\b	(?{ $_{suffix} = $^N }))?
-          )
-        )
-	/ix;
-
-    # http://www.usps.com/ncsc/lookups/abbreviations.html#secunitdesig
-    # TODO add support for those that don't require a number
-    # TODO map to standard names/abbreviations
-    $Addr_Match{sec_unit_type_numbered} = qr/
-          (su?i?te
-            |p\W*[om]\W*b(?:ox)?
-            |(?:ap|dep)(?:ar)?t(?:me?nt)?
-            |ro*m
-            |flo*r?
-            |uni?t
-            |bu?i?ldi?n?g
-            |ha?nga?r
-            |lot
-            |pier
-            |slip
-            |spa?ce?
-            |stop
-            |tra?i?le?r
-            |box)\b				(?{ $_{sec_unit_type}   = $^N })
-        /ix;
-
-    $Addr_Match{sec_unit_type_unnumbered} = qr/
-          (ba?se?me?n?t
-            |fro?nt
-            |lo?bby
-            |lowe?r
-            |off?i?ce?
-            |pe?n?t?ho?u?s?e?
-            |rear
-            |side
-            |uppe?r
-            )\b			        	(?{ $_{sec_unit_type}   = $^N })
-        /ix;
-
-    $Addr_Match{sec_unit} = qr/
-        (:?
-            (?: (?:$Addr_Match{sec_unit_type_numbered} \W+)
-                | (\#)\W*                       (?{ $_{sec_unit_type}   = $^N })
-            )
-            (  [\w-]+)				(?{ $_{sec_unit_num}    = $^N })
-        )
-        |
-            $Addr_Match{sec_unit_type_unnumbered}
-        /ix;
-
-    $Addr_Match{city_and_state} = qr/
-	(?:
-	    ([^\d,]+?)\W+			(?{ $_{city}   = $^N })
-	    ($Addr_Match{state})		(?{ $_{state}  = $^N })
-	)
-	/ix;
-
-    $Addr_Match{place} = qr/
-	(?:$Addr_Match{city_and_state}\W*)?
-	(?:($Addr_Match{zip})			(?{ $_{zip}    = $^N }))?
-	/ix;
-
-    $Addr_Match{address} = qr/^\W*
-	(  $Addr_Match{number})\W*		(?{ $_{number} = $^N })
-        (?:$Addr_Match{fraction}\W*)?
-	   $Addr_Match{street}\W+
-	(?:$Addr_Match{sec_unit}\W+)?
-	   $Addr_Match{place}
-	\W*$/ix;
-
-    $Addr_Match{informal_address} = qr/
-        ^\s*
-        (?:$Addr_Match{sec_unit}\W+)?
-        (  $Addr_Match{number})?\W*             (?{ $_{number} = $^N })
-        (?:$Addr_Match{fraction}\W*)?
-           $Addr_Match{street}\W+
-        (?:$Addr_Match{sec_unit}\W+)?
-        (?:$Addr_Match{place})?
-        /ix;
-
-    $Addr_Match{intersection} = qr/^\W*
-	   $Addr_Match{street}\W*?	
-
-	\s+$Addr_Match{corner}\s+
-
-	    (?{ exists $_{$_} and $_{$_.1} = delete $_{$_} for (qw{prefix street type suffix})})
-	   $Addr_Match{street}\W+
-	    (?{ exists $_{$_} and $_{$_.2} = delete $_{$_} for (qw{prefix street type suffix})})
-
-	   $Addr_Match{place}
-	\W*$/ix;
-}
+init();
 
 our %Normalize_Map = (
     prefix  => \%Directional,
@@ -854,6 +723,180 @@ our $Old_Undef_Fields_Behaviour = 0;
 
 =head1 CLASS METHODS
 
+=head2 init
+
+    # Add another street type mapping:
+    $Geo::StreetAddress::US::Street_Type{'cur'}='curv';
+    # Re-initialize to pick up the change
+    Geo::StreetAddress::US::init();
+
+Runs the setup on globals.  This is run automatically when the module is loaded,
+but if you subsequently change the globals, you should run it again.
+
+=cut
+
+sub init {
+
+    %Direction_Code = reverse %Directional;
+
+    %_Street_Type_List = map { $_ => 1 } %Street_Type;
+
+    %FIPS_State = reverse %State_FIPS;
+
+    use re 'eval';
+
+    %Addr_Match = (
+        type    => join("|", keys %_Street_Type_List),
+        fraction => qr{\d+\/\d+},
+        state   => '\b(?:'.join("|",
+   		    # escape spaces in state names (e.g., "new york" --> "new\\ york")
+	        # so they still match in the x environment below
+            map { ( quotemeta $_) } keys %State_Code, values %State_Code
+            ).')\b',
+        direct  => join("|",
+			    # map direction names to direction codes
+                        keys %Directional,
+			    # also map the dotted version of the code to the code itself
+                        map { my $c = $_;
+                              $c =~ s/(\w)/$1./g;
+                              ( quotemeta $c, $_ ) }
+                        sort { length $b <=> length $a }
+                        values %Directional),
+        dircode => join("|", keys %Direction_Code),
+        zip	    => qr/\d{5}(?:-?\d{4})?/,  # XXX add \b?
+        corner  => qr/(?:\band\b|\bat\b|&|\@)/i,
+    );
+
+    # we don't include letters in the number regex because we want to
+    # treat "42S" as "42 S" (42 South). For example,
+    # Utah and Wisconsin have a more elaborate system of block numbering
+    # http://en.wikipedia.org/wiki/House_number#Block_numbers
+    $Addr_Match{number} = qr/(\d+-?\d*) (?{ $_{number} = $^N })/ix,
+
+    # note that expressions like [^,]+ may scan more than you expect
+    $Addr_Match{street} = qr/
+        (?:
+          # special case for addresses like 100 South Street
+          (?:($Addr_Match{direct})\W+           (?{ $_{street} = $^N })
+             ($Addr_Match{type})\b              (?{ $_{type}   = $^N }))
+             #(?{ $_{_street}.=1 })
+          |
+          (?:($Addr_Match{direct})\W+		(?{ $_{prefix} = $^N }))?
+          (?:
+            ([^,]*\d)				(?{ $_{street} = $^N })
+            (?:[^\w,]*($Addr_Match{direct})\b	(?{ $_{suffix} = $^N; $_{type}||='' }))
+            #(?{ $_{_street}.=3 })
+           |
+            ([^,]+)				(?{ $_{street} = $^N })
+            (?:[^\w,]+($Addr_Match{type})\b	(?{ $_{type}   = $^N }))
+            (?:[^\w,]+($Addr_Match{direct})\b	(?{ $_{suffix} = $^N }))?
+            #(?{ $_{_street}.=2 })
+           |
+            ([^,]+?)				(?{ $_{street} = $^N; $_{type}||='' })
+            (?:[^\w,]+($Addr_Match{type})\b	(?{ $_{type}   = $^N }))?
+            (?:[^\w,]+($Addr_Match{direct})\b	(?{ $_{suffix} = $^N }))?
+            #(?{ $_{_street}.=4 })
+          )
+        )
+	/ix;
+
+
+    # http://pe.usps.com/text/pub28/pub28c2_003.htm
+    # TODO add support for those that don't require a number
+    # TODO map to standard names/abbreviations
+    $Addr_Match{sec_unit_type_numbered} = qr/
+          (su?i?te
+            |p\W*[om]\W*b(?:ox)?
+            |(?:ap|dep)(?:ar)?t(?:me?nt)?
+            |ro*m
+            |flo*r?
+            |uni?t
+            |bu?i?ldi?n?g
+            |ha?nga?r
+            |lo?t
+            |pier
+            |slip
+            |spa?ce?
+            |stop
+            |tra?i?le?r
+            |box)(?![a-z])			(?{ $_{sec_unit_type}   = $^N })
+        /ix;
+
+    $Addr_Match{sec_unit_type_unnumbered} = qr/
+          (ba?se?me?n?t
+            |fro?nt
+            |lo?bby
+            |lowe?r
+            |off?i?ce?
+            |pe?n?t?ho?u?s?e?
+            |rear
+            |side
+            |uppe?r
+            )\b			        	(?{ $_{sec_unit_type}   = $^N })
+        /ix;
+
+    $Addr_Match{sec_unit} = qr/
+        (:?
+            (?: (?:$Addr_Match{sec_unit_type_numbered} \W*)
+                | (\#)\W*                       (?{ $_{sec_unit_type}   = $^N })
+            )
+            (  [\w-]+)				(?{ $_{sec_unit_num}    = $^N })
+        )
+        |
+            $Addr_Match{sec_unit_type_unnumbered}
+        /ix;
+
+    $Addr_Match{city_and_state} = qr/
+	(?:
+	    ([^\d,]+?)\W+			(?{ $_{city}   = $^N })
+	    ($Addr_Match{state})		(?{ $_{state}  = $^N })
+	)
+	/ix;
+
+    $Addr_Match{place} = qr/
+	(?:$Addr_Match{city_and_state}\W*)?
+	(?:($Addr_Match{zip})			(?{ $_{zip}    = $^N }))?
+	/ix;
+
+    $Addr_Match{address} = qr/
+        ^
+        [^\w\#]*    # skip non-word chars except # (eg unit)
+	(  $Addr_Match{number} )\W*
+        (?:$Addr_Match{fraction}\W*)?
+	   $Addr_Match{street}\W+
+	(?:$Addr_Match{sec_unit}\W+)?
+	   $Addr_Match{place}
+	\W*         # require on non-word chars at end
+        $           # right up to end of string
+        /ix;
+
+    my $sep = qr/(?:\W+|\Z)/;
+
+    $Addr_Match{informal_address} = qr/
+        ^
+        \s*         # skip leading whitespace
+        (?:$Addr_Match{sec_unit} $sep)?
+        (?:$Addr_Match{number})?\W*
+        (?:$Addr_Match{fraction}\W*)?
+           $Addr_Match{street} $sep
+        (?:$Addr_Match{sec_unit} $sep)?
+        (?:$Addr_Match{place})?
+        # don't require match to reach end of string
+        /ix;
+
+    $Addr_Match{intersection} = qr/^\W*
+	   $Addr_Match{street}\W*?	
+
+	\s+$Addr_Match{corner}\s+
+
+	    (?{ exists $_{$_} and $_{$_.1} = delete $_{$_} for (qw{prefix street type suffix})})
+	   $Addr_Match{street}\W+
+	    (?{ exists $_{$_} and $_{$_.2} = delete $_{$_} for (qw{prefix street type suffix})})
+
+	   $Addr_Match{place}
+	\W*$/ix;
+}
+
 =head2 parse_location
 
     $spec = Geo::StreetAddress::US->parse_location( $string )
@@ -979,9 +1022,11 @@ is a problem and I'll see what I can do to fix it.
 
 sub normalize_address {
     my ($class, $part) = @_;
-    
-    # strip off punctuation
-    defined($_) && s/^\s+|\s+$|[^\w\s\-]//gos for values %$part;
+
+    #m/^_/ and delete $part->{$_} for keys %$part; # for debug
+
+    # strip off some punctuation
+    defined($_) && s/^\s+|\s+$|[^\w\s\-\#\&]//gos for values %$part;
 
     if ($Old_Undef_Fields_Behaviour) {
         my @undef_fields = (exists $part->{street1})
@@ -996,7 +1041,7 @@ sub normalize_address {
 	      and exists $map->{lc $part->{$key}};
     }
 
-    $part->{$_} = ucfirst lc $part->{$_} 
+    $part->{$_} = ucfirst lc $part->{$_}
 	for grep(exists $part->{$_}, qw( type type1 type2 ));
 
     # attempt to expand directional prefixes on place names
@@ -1029,6 +1074,12 @@ for your purposes. If this turns out to be the case, please let me know.
 
 Geo::StreetAddress::US does B<NOT> perform USPS-certified address normalization.
 
+Grid based addresses, like those in Utah, where the direction comes before the
+number, e.g. W164N5108 instead of 164 W 5108 N, aren't handled at the moment.
+A workaround is to apply a regex like this
+
+    s/([nsew])\s*(\d+)\s*([nsew])\s*(\d+)/$2 $1 $4 $3/
+
 =head1 SEE ALSO
 
 This software was originally part of Geo::Coder::US(3pm).
